/*	$AoiMoe: bmp.cc,v 1.2 2001/08/08 19:24:23 tshiozak Exp $	*/

/*
 * Copyright (c) 2001 Takuya SHIOZAKI <AoiMoe@imou.to>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */


#include "types.h"
#include "debug.h"
#include "vinfo.h"
#include "list.h"
#include "rectset.h"
#include "canvas.h"
#include "canvas_mi_p.h"
#include "event.h"
#include "timeout.h"
#include "servant.h"
#include "lookup.h"
#include "file.h"

#pragma implementation "bmp.h"

#include "bmp.h"

class BMPCanvas : public SimpleMICanvas {
protected:
	struct BitmapFileHeader {
//		uint16_t	bfType; 
		uint32_t	bfSize; 
//		uint16_t	bfReserved1; 
//		uint16_t	bfReserved2; 
		uint32_t	bfOffBits; 
	};
	struct BitmapInfoHeader {
		uint32_t	biSize;
		int32_t		biWidth;
		int32_t		biHeight;
		int16_t		biPlanes;
		int16_t		biBitCount;
		uint32_t	biCompression;
		uint32_t	biSizeImage;
		int32_t		biXPelsPerMeter;
		int32_t		biYPelsPerMeter;
		uint32_t	biClrUsed;
		uint32_t	biClrImportant;
	};
	static size_t const lenBitmapFileHeader = 14;
	static size_t const lenBitmapInfoHeader = 40;
	static bool get_fileheader(BitmapFileHeader *fh, IFilePtr &fp)
	{
		bool ret = true;
		if (fp.remain() < lenBitmapFileHeader) {
			AMUSE_DPR((
				"BMP: cannot get fileheader (too short)\n"));
			goto quit;
		}
		if (fp.get8() != 'B') {
			AMUSE_DPR((
				"BMP: cannot get fileheader "
				"(unmatch magic)\n"));
			goto quit;
		}
		if (fp.get8() != 'M') {
			AMUSE_DPR((
				"BMP: cannot get fileheader "
				"(unmatch magic)\n"));
			goto quit;
		}
		fh->bfSize = fp.get32le();
		fp.get16le();
		fp.get16le();
		fh->bfOffBits = fp.get32le();
		ret = false;
	quit:
		return ret;
	}
	static bool get_infoheader(BitmapInfoHeader *ih, IFilePtr &fp)
	{
		bool ret = true;
		size_t remain;
		if (fp.remain() < lenBitmapInfoHeader) {
			AMUSE_DPR((
				"BMP: cannot get infoheader (too short)\n"));
			goto quit;
		}
		remain = ih->biSize = fp.get32le(); remain-=4;
		if (ih->biSize < lenBitmapInfoHeader) {
			AMUSE_DPR((
				"BMP: cannot get infoheader (too short)\n"));
			goto quit;
		}
		ih->biWidth = fp.get32le(); remain-=4;
		ih->biHeight = fp.get32le(); remain-=4;
		ih->biPlanes = fp.get16le(); remain-=2;
		ih->biBitCount = fp.get16le(); remain-=2;
		ih->biCompression = fp.get32le(); remain-=4;
		ih->biSizeImage = fp.get32le(); remain-=4;
		ih->biXPelsPerMeter = fp.get32le(); remain-=4;
		ih->biYPelsPerMeter = fp.get32le(); remain-=4;
		ih->biClrUsed = fp.get32le(); remain-=4;
		ih->biClrImportant = fp.get32le(); remain-=4;
		if (remain>0)
			fp.fwd(remain);
		ret = false;
	quit:
		return ret;
	}
	struct MatchParam {
		char const *name_;
		BitmapFileHeader const *fileHeader_;
		BitmapInfoHeader const *infoHeader_;
		MatchParam(char const *n,
			   BitmapFileHeader const *fh,
			   BitmapInfoHeader const *ih) :
			name_(n), fileHeader_(fh), infoHeader_(ih)
		{
		}
	};
	typedef bool (*MatchFunc)(MatchParam const &);
	typedef Canvas *(*FactoryFunc)(MatchParam const &, IFilePtr &fp);
	struct MatchContext {
		MatchFunc match;
		FactoryFunc factory;
	};
	static MatchContext const *matchContextList[];
public:
	~BMPCanvas() { }
	BMPCanvas(WH const &wh) : SimpleMICanvas(wh) { }
	static Canvas *create(char const *path)
	{
		Canvas *cv = NULL;
		IFile *file = new IFile(path);
		MatchContext const **p;
		BitmapFileHeader fh;
		BitmapInfoHeader ih;

		if (!file || !*file) {
			AMUSE_DPR(("failed to open file %s\n", path));
			goto quit;
		}
		{
			IFilePtr fp(file);
			if (get_fileheader(&fh, fp))
				goto quit;
			if (get_infoheader(&ih, fp))
				goto quit;
			if (ih.biSize > fh.bfOffBits) {
				AMUSE_DPR(("inconsist biSize(%d) "
					   "and bfOffBits(%d)\n",
					   ih.biSize, fh.bfOffBits));
				goto quit;
			}
			for (p=matchContextList; *p; p++) {
				if ((*p)->match(MatchParam(path, &fh, &ih)))
					break;
			}
			if (*p) {
				cv = (*p)->factory(
					MatchParam(path, &fh, &ih), fp);
			} else {
				AMUSE_DPR(("BMP: cannot match\n"));
			}
		}
	quit:
		if (file)
			delete file;
		return cv;
	}
};


// ----------------------------------------------------------------------

class DirectBMPCanvas24 : public BMPCanvas {
	bool read_file(IFilePtr &fp,
		       BitmapFileHeader const *fh, BitmapInfoHeader const *ih)
	{
		MIPixel *pixel, *line;
		RGBA<uint8_t> tmp;
		int delta, count, x, y;

		if (fp.remain()+ih->biSize+lenBitmapFileHeader<
		    fh->bfOffBits+wh().w()*wh().h()*3) {
			AMUSE_DPR(("BMP: too short file body.\n"));
			return true;
		}
		fp.fwd(fh->bfOffBits-ih->biSize-lenBitmapFileHeader);

		line = (MIPixel *)get_frame_raw();
		if (ih->biHeight>0) {
			line += (ih->biHeight-1)*ih->biWidth;
			delta = -ih->biWidth;
			count = ih->biHeight;
		} else {
			delta = ih->biWidth;
			count = -ih->biHeight;
		}
		for (y=count; y>0; y--) {
			pixel = line;
			for (x=ih->biWidth; x>0; x--) {
				tmp.b() = fp.get8();
				tmp.g() = fp.get8();
				tmp.r() = fp.get8();
				tmp.a() = MIPixel::intensityMaxA;
				*pixel = MIPixel(tmp.r(),
						 tmp.b(),
						 tmp.g(),
						 tmp.a());
				pixel++;
			}
			line += delta;
		}
		add_clip(Rect(Pos(0, 0), wh()));
		add_dirty(Rect(Pos(0, 0), wh()));

		return false;
	}
public:
	DirectBMPCanvas24(WH const &wh) : BMPCanvas(wh) { }
	static bool match(MatchParam const &mp)
	{
		if (mp.infoHeader_->biBitCount == 24)
			return true;

		return false;
	}
	static Canvas *factory(MatchParam const &mp, IFilePtr &fp)
	{
		DirectBMPCanvas24 *bmp =
		    new DirectBMPCanvas24(
			    WH(mp.infoHeader_->biWidth,
			       mp.infoHeader_->biHeight>0?
			       mp.infoHeader_->biHeight:
			       -mp.infoHeader_->biHeight));

		if (bmp->read_file(fp, mp.fileHeader_, mp.infoHeader_)) {
			delete bmp;
			bmp = NULL;
		}

		return bmp;
	}
	static MatchContext const matchContext;
};

BMPCanvas::MatchContext const DirectBMPCanvas24::matchContext = {
	&match, &factory
};


// ----------------------------------------------------------------------

BMPCanvas::MatchContext const *BMPCanvas::matchContextList[] = {
	&DirectBMPCanvas24::matchContext,
	NULL
};


// ----------------------------------------------------------------------

Canvas *BMPCanvasFactory::create(char const *file)
{
	return BMPCanvas::create(file);
}
