/*	$AoiMoe: canvas.cc,v 1.15 2001/08/08 19:28:12 tshiozak Exp $	*/

/*
 * Copyright (c) 2001 Takuya SHIOZAKI <AoiMoe@imou.to>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */


#include <sys/types.h>
#include <string.h>
#include "types.h"
#include "debug.h"
#include "vinfo.h"
#include "list.h"
#include "rectset.h"
#include "canvas.h"
#include "canvas_mi_p.h"
#include "clip.h"

// ----------------------------------------------------------------------

class MIVisualInfo : public VisualInfo {
public:
	MIVisualInfo()
	{
		bits_per_pixel_ = sizeof(MIPixel)*8;
		depth_ = bits_per_pixel_;
		scanline_pad_ = bits_per_pixel_;
		hasAlpha_ = true;
		endian_ = htons(1)==1 ? Big:Little;
		shifts_.r().lshift = MIPixel::shiftR;
		shifts_.g().lshift = MIPixel::shiftG;
		shifts_.b().lshift = MIPixel::shiftB;
		shifts_.a().lshift = MIPixel::shiftA;
		shifts_.r().rshift = 8-MIPixel::bitsR;
		shifts_.g().rshift = 8-MIPixel::bitsG;
		shifts_.b().rshift = 8-MIPixel::bitsB;
		shifts_.a().rshift = 8-MIPixel::bitsA;
	}
	static MIVisualInfo const miVisualInfo;
};

MIVisualInfo const MIVisualInfo::miVisualInfo;
VisualInfo const &MICanvas::miVisualInfo = MIVisualInfo::miVisualInfo;

void Canvas::sync(bool isclear) const
{
	for (CanvasRegion::Iterator i=get_dirty(); i; ++i) {
		Rect r;
		conv_rect(&r, i);
		notify_update(r);
	}
	if (isclear)
		clear_dirty();
	notify_sync(isclear);
}

// ----------------------------------------------------------------------


SimpleMICanvas::SimpleMICanvas(WH const &wh) :
	MICanvas(wh),
	frame_(new MIPixel[wh.w()*wh.h()])
{
	bytes_per_line_ = wh.w()*visual_info().bits_per_pixel()/8;
}
SimpleMICanvas::~SimpleMICanvas() { delete [] frame_; }
void *SimpleMICanvas::get_frame_raw() const { return (void *)frame_; }

MICanvas *MICanvasFactory::create(WH const &wh)
{
	MICanvas *canvas;

	canvas = new SimpleMICanvas(wh);
	if (canvas && !*canvas) {
		delete canvas;
		canvas = NULL;
	}

	return canvas;
}


// ----------------------------------------------------------------------

class SimpleMDCanvas : public MDCanvas, public SimpleCanvasRegionImpl {
	void *frame_;
public:
	SimpleMDCanvas(VisualInfo const &vi, CMap const &cmap, WH const wh) :
		MDCanvas(vi, cmap, wh),
		frame_(new uint8_t [wh.w()*wh.h()])
	{
	}
	~SimpleMDCanvas() { delete (uint8_t *)frame_; }
	void *get_frame_raw() const { return frame_; }
};

MDCanvas *
MDCanvasFactory::create(VisualInfo const &vi, CMap const &cmap, WH const wh)
{
	MDCanvas *canvas;

	canvas = new SimpleMDCanvas(vi, cmap, wh);
	if (canvas && !*canvas) {
		delete canvas;
		canvas = NULL;
	}

	return canvas;
}

// ----------------------------------------------------------------------

ViewPortCanvas::ViewPortCanvas(Canvas const &c, Rect const &rect) :
	Canvas(c.visual_info())
{
	bind_canvas(c, rect);
}

void ViewPortCanvas::bind_canvas(Canvas const &c, Rect const &rect)
{
	Rect r(c.wh());
	pos_ = rect.to_pos();
	r &= rect;
	wh_ = r.to_wh();
	canvas_ = &c;
	bytes_per_line_ = c.bytes_per_line();
	vinfo_ = &canvas_->visual_info();
	canvas_->update_notifier().connect(this);
	canvas_->sync_notifier().connect(this);
}

void *ViewPortCanvas::get_frame_raw() const
{
	return (void *)((uint8_t *)canvas_->get_frame_raw()+
			pos_.x()*(visual_info().bits_per_pixel()/8)+
			pos_.y()*bytes_per_line_);
}

void ViewPortCanvas::sync(bool isclear) const
{
	canvas_->sync(isclear);
}

Canvas::Kind ViewPortCanvas::kind() const
{
	return canvas_->kind();
}

void ViewPortCanvas::add_clip(Rect const &d) const
{
	Rect tmp(d);
	tmp += pos_;
	canvas_->add_clip(tmp);
}

void ViewPortCanvas::clear_clip() const
{
	canvas_->clear_clip();
}

CanvasRegion::Iterator ViewPortCanvas::get_clip() const
{
	return canvas_->get_clip();
}

void ViewPortCanvas::add_dirty(Rect const &d) const
{
	Rect tmp(d);
	tmp += pos_;
	canvas_->add_dirty(tmp);
}

void ViewPortCanvas::clear_dirty() const
{
	canvas_->clear_dirty();
}

CanvasRegion::Iterator ViewPortCanvas::get_dirty() const
{
	return canvas_->get_dirty();
}

void ViewPortCanvas::conv_rect(Rect *r, Iterator const &i) const
{
	*r = *to_rect_iter(i);
	*r -= pos_;
	*r &= Rect(Pos(0, 0), wh_);
}

CMap const &ViewPortCanvas::cmap() const
{
	return canvas_->cmap();
}

void ViewPortCanvas::callback(CanvasUpdateNotifier const &,
			      Canvas const &cv, Rect const &rect)
{
	AMUSE_ASSERT(&cv == canvas_);

	Rect r(rect);
	r -= pos_;
	r &= Rect(Pos(0, 0), wh_);
	if (r.w())
		notify_update(r);
	canvas_->update_notifier().connect(this);
}

void ViewPortCanvas::callback(CanvasSyncNotifier const &,
			      Canvas const &cv, bool isclear)
{
	notify_sync(isclear);
	canvas_->sync_notifier().connect(this);
}


// ----------------------------------------------------------------------

typedef void (*BltFunc)(Canvas const &, Pos const &,
			Canvas const &, Pos const &, WH const &, bool);

void
bitblt_homo(Canvas const &dst, Pos const &dstpos,
	    Canvas const &src, Pos const &srcpos, WH const &wh, bool key)
{
	uint8_t *psrc, *pdst;
	size_t linew, srclw, dstlw, bpp, skipsrc, skipdst;
	int x, y;

	AMUSE_ASSERT(dst.visual_info() == src.visual_info());
	srclw = src.bytes_per_line();
	dstlw = dst.bytes_per_line();
	bpp = src.visual_info().bits_per_pixel()/8;
	psrc =
	    (uint8_t *)src.get_frame_raw() +
	    srcpos.x()*bpp + srcpos.y()*srclw;
	pdst =
	    (uint8_t *)dst.get_frame_raw() +
	    dstpos.x()*bpp + dstpos.y()*dstlw;
	linew = wh.w()*bpp;

	if (key) {
#define BLT_(cmp, store, incr) \
		for (y=wh.h(); y>0; y--) { \
			for (x=wh.w(); x>0; x--) { \
				if (cmp) { \
					store; \
				} \
				incr; \
			} \
			psrc+=skipsrc; pdst+=skipdst; \
		}
		AMUSE_ASSERT(srclw>=linew && dstlw>=linew);
		skipsrc = srclw-linew;
		skipdst = dstlw-linew;
		if (src.kind() == Canvas::MI) {
			BLT_((!((MIPixel *)psrc)->iskey()),
			     (*(MIPixel *)pdst = *(MIPixel *)psrc),
			     (psrc+=sizeof(MIPixel),pdst+=sizeof(MIPixel)));
		} else {
			AMUSE_DPR(("key bitblt is not implemented "
				   "for MD canvas\n"));
			goto XXX;
		}
#undef BLT_
	} else {
XXX:
		for (y=wh.h(); y>0; y--) {
			memmove(pdst, psrc, linew);
			psrc += srclw; pdst += dstlw;
		}
	}
}

void
bitblt_mdmi(Canvas const &dst, Pos const &dstpos,
	    Canvas const &src, Pos const &srcpos, WH const &wh, bool key)
{
	CMap const &cmap = static_cast<MDCanvas const &>(dst).cmap();
	int x, y;
	size_t srclw, srcskip, dstlw, dstskip, dstbpp, srcbpp;
	uint8_t *psrc, *pdst;
	CMap::Cell v;

	srcbpp = src.visual_info().bits_per_pixel()/8;
	dstbpp = dst.visual_info().bits_per_pixel()/8;
	srclw = src.bytes_per_line();
	dstlw = dst.bytes_per_line();
	psrc =
	    (uint8_t *)src.get_frame_raw() +
	    srcpos.x()*srcbpp + srcpos.y()*srclw;
	pdst =
	    (uint8_t *)dst.get_frame_raw() +
	    dstpos.x()*dstbpp + dstpos.y()*dstlw;
	AMUSE_ASSERT(srclw>=wh.w()*srcbpp);
	AMUSE_ASSERT(dstlw>=wh.w()*dstbpp);
	srcskip = srclw - wh.w()*srcbpp;
	dstskip = dstlw - wh.w()*dstbpp;
#define BLT_(cmp) \
	for (y=wh.h(); y>0; y--) { \
		for (x=wh.w(); x>0; x--) { \
			if (cmp) { \
				v = cmap.r()[((MIPixel *)psrc)->r()] | \
				    cmap.g()[((MIPixel *)psrc)->g()] | \
				    cmap.b()[((MIPixel *)psrc)->b()]; \
				memcpy(pdst, &v, dstbpp); \
			} \
			psrc += srcbpp; \
			pdst += dstbpp; \
		} \
		psrc+=srcskip; \
		pdst+=dstskip; \
	}
	if (key) {
		BLT_(!((MIPixel *)psrc)->iskey());
	} else {
		BLT_(true);
	}
#undef BLT_
}
void
bitblt_mimd(Canvas const &, Pos const &,
	    Canvas const &, Pos const &, WH const &, bool)
{
	AMUSE_ASSERT(!"not implemented");
}

void
bitblt(Canvas const &dst, Pos const &dstpos,
       Canvas const &src, Pos const &srcpos, WH const &wh,
       bool key, bool usedirty)
{
	BltFunc blt;
	CanvasRegion::Iterator i;
	RectSet rset;

	static BltFunc const tbl[2][2] = {
		{ bitblt_homo, bitblt_mdmi },
		{ bitblt_mimd, bitblt_homo }
	};

	AMUSE_ASSERT(Canvas::MD==0 && Canvas::MI==1 &&
		     (dst.kind() == Canvas::MD ||
		      dst.kind() == Canvas::MI) &&
		     (src.kind() == Canvas::MD ||
		      src.kind() == Canvas::MI));

	blt = tbl[dst.kind()][src.kind()];

	if (usedirty)
		i = CanvasRegion::Iterator(src.get_dirty());
	else {
		i = CanvasRegion::Iterator(src.get_clip());
	}

	for (; i; ++i) {
		Rect dstrect(Pos(0, 0), dst.wh());
		Rect srcrect;
		src.conv_rect(&srcrect, i);
		if (clip(&dstrect, &srcrect, dstpos, srcpos, wh)) {
			dst.add_clip(dstrect);
			dst.add_dirty(dstrect);
			blt(dst, dstrect.to_pos(),
			    src, srcrect.to_pos(), srcrect.to_wh(), key);
		}
#ifdef AMUSE_DEBUG
		 else {
			Rect r;
			src.conv_rect(&r, i);
			AMUSE_DPR(("bitblt: clipped.  src=%dx%d+%d+%d\n",
				   r.w(), r.h(),
				   r.x(), r.y()));
		}
#endif
	}
}


// ----------------------------------------------------------------------

void
fill(Canvas const &c, Rect const &rect, void const *pix)
{
	uint8_t *frame;
	size_t skip;
	size_t bpp;
	Rect r(Pos(0, 0), c.wh());
	int x, y;

	r &= rect;
	if (r.w()) {
		c.add_clip(r);
		c.add_dirty(r);
		bpp = c.visual_info().bits_per_pixel()/8;
		skip = c.bytes_per_line() - r.w()*bpp;
		frame =
		    (uint8_t *)c.get_frame_raw() +
		    r.x()*bpp + r.y()*c.bytes_per_line();

		for (y=r.h(); y>0; y--) {
			for (x=r.w(); x>0; x--) {
				memcpy(frame, pix, bpp);
				frame += bpp;
			}
			frame += skip;
		}
	}
}

void
fill(Canvas const &c, Rect const &rect, MIPixel const &pix)
{
	if (c.kind() == Canvas::MI)
		fill(c, rect, (void const *)&pix);
	else {
		CMap const &cmap = static_cast<MDCanvas const &>(c).cmap();
		uint32_t v;
		v = cmap.r()[pix.r()] |
		    cmap.g()[pix.g()] |
		    cmap.b()[pix.b()];
		fill(c, rect, (void const *)&v);
	}
}
