/*	$AoiMoe: canvas.h,v 1.16 2001/08/08 19:28:12 tshiozak Exp $	*/

/*
 * Copyright (c) 2001 Takuya SHIOZAKI <AoiMoe@imou.to>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */


#ifndef AMUSE_CANVAS_H_
#define AMUSE_CANVAS_H_

#include "list.h" // XXX
#include "event.h" // XXX

class Canvas;
AM_DEF_SIMPLE_RECEIVERNOTIFIER(CanvasUpdate,
			       (CanvasUpdateNotifier const &notifier,
				Canvas const &canvas, Rect const &rect),
			       (notifier, canvas, rect));
AM_DEF_SIMPLE_RECEIVERNOTIFIER(CanvasSync,
			       (CanvasSyncNotifier const &notifier,
				Canvas const &canvas, bool cleardirty),
			       (notifier, canvas, cleardirty));

class CanvasRegion {
	class Iterator {
		friend Canvas;
		friend CanvasRegion;
		RectSet::Iterator iter_;
	public:
		explicit Iterator(RectSet const &r) :
			iter_(r)
		{
		}
		Iterator() { }
		operator bool () const
		{
			return iter_;
		}
		void operator ++ ()
		{
			++iter_;
		}
		void operator -- ()
		{
			--iter_;
		}
	};
protected:
	static RectSet::Iterator const &to_rect_iter(Iterator const &i)
	{
		return i.iter_;
	}
public:
	virtual void add_clip(Rect const &) const = 0;
	virtual void clear_clip() const = 0;
	virtual Iterator get_clip() const = 0;
	virtual void add_dirty(Rect const &) const = 0;
	virtual void clear_dirty() const = 0;
	virtual Iterator get_dirty() const = 0;
	virtual void conv_rect(Rect *, Iterator const &) const = 0;
};

class SimpleCanvasRegionImpl : virtual public CanvasRegion {
	mutable RectSet clip_;
	mutable RectSet dirty_;
public:
	virtual void add_clip(Rect const &d) const { clip_ |= d; }
	virtual void clear_clip() const { clip_.clear(); }
	virtual Iterator get_clip() const { return Iterator(clip_); }
	virtual void add_dirty(Rect const &d) const { dirty_ |= d; }
	virtual void clear_dirty() const { dirty_.clear(); }
	virtual Iterator get_dirty() const { return Iterator(dirty_); }
	virtual void conv_rect(Rect *r, Iterator const &i) const
	{
		*r = *to_rect_iter(i);
	}
};

class Canvas : virtual public CanvasRegion {
public:
	enum Kind { MD=0, MI=1 };
protected:
	mutable SimpleCanvasUpdateNotifier canvasUpdateNotifier_;
	mutable SimpleCanvasSyncNotifier canvasSyncNotifier_;
	VisualInfo const *vinfo_;
	WH wh_;
	unsigned int bytes_per_line_;
	Canvas(VisualInfo const &vi, WH const &wh=WH(0,0)) :
		vinfo_(&vi), wh_(wh)
	{
	}
	Canvas() : vinfo_(NULL) { }
public:
	virtual ~Canvas() { }
	WH const &wh() const { return wh_; }
	VisualInfo const &visual_info() const { return *vinfo_; }
	unsigned int bytes_per_line() const { return bytes_per_line_; }
	operator bool () const { return !!get_frame_raw(); }
	CanvasUpdateNotifier &update_notifier() const
	{
		return canvasUpdateNotifier_;
	}
	CanvasSyncNotifier &sync_notifier() const
	{
		return canvasSyncNotifier_;
	}
	//
	virtual Kind kind() const = 0;
	virtual void *get_frame_raw() const = 0;
	virtual void sync(bool =true) const;
	virtual CMap const &cmap() const = 0;
	void notify_update(Rect const &rect) const
	{
		canvasUpdateNotifier_.notify(
			canvasUpdateNotifier_, *this, rect);
	}
	void notify_sync(bool cleardirty) const
	{
		canvasSyncNotifier_.notify(canvasSyncNotifier_,
					   *this, cleardirty);
	}
};

class MDCanvas : public Canvas {
protected:
	CMap const &cmap_;
	MDCanvas(VisualInfo const &vi, CMap const &cmap, WH const &wh) :
		Canvas(vi, wh), cmap_(cmap)
	{
	}
public:
	virtual CMap const &cmap() const { return cmap_; }
	virtual Kind kind() const { return MD; }
};

namespace MDCanvasFactory {
	MDCanvas *create(VisualInfo const &vi, CMap const &cmap, WH const wh);
}


class MICanvas : public Canvas {
	static VisualInfo const &miVisualInfo;
protected:
	MICanvas(WH const &wh=WH(0,0)) : Canvas(miVisualInfo, wh) { }
public:
	virtual Kind kind() const { return MI; }
	virtual MIPixel *get_frame() const
	{
		return (MIPixel *)get_frame_raw();
	}
	virtual CMap const &cmap() const
	{
		return *(CMap *)0;
	}
};

namespace MICanvasFactory {
	MICanvas *create(WH const &wh);
}


class ViewPortCanvas : public Canvas,
		       public CanvasUpdateReceiver,
		       public CanvasSyncReceiver {
	Canvas const *canvas_;
	Pos pos_;
public:
	ViewPortCanvas() : Canvas(), canvas_(0) { }
	ViewPortCanvas(Canvas const &, Rect const &);
	void bind_canvas(Canvas const &, Rect const &);
	//
	virtual void *get_frame_raw() const;
	virtual void sync(bool) const;
	virtual Kind kind() const;
	//
	virtual void add_clip(Rect const &) const;
	virtual void clear_clip() const;
	virtual Iterator get_clip() const;
	virtual void add_dirty(Rect const &) const;
	virtual void clear_dirty() const;
	virtual Iterator get_dirty() const;
	virtual void conv_rect(Rect *, Iterator const &) const;
	virtual CMap const &cmap() const;
	//
protected:
	virtual void callback(CanvasUpdateNotifier const &,
			      Canvas const &, Rect const &);
	virtual void callback(CanvasSyncNotifier const &,
			      Canvas const &, bool);
};

void
bitblt(Canvas const &/*dst*/, Pos const &/*dstpos*/,
       Canvas const &/*src*/, Pos const &/*srcpos*/, WH const &/*wh*/,
       bool /*key*/, bool =false /*usedirty*/);

void fill(Canvas const &, Rect const &, void const *);
void fill(Canvas const &, Rect const &, MIPixel const &);


#endif
