/*	$AoiMoe: sss_ini.cc,v 1.2 2001/08/08 19:40:30 tshiozak Exp $	*/

/*
 * Copyright (c) 2001 Takuya SHIOZAKI <AoiMoe@imou.to>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */


#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include "types.h"
#include "debug.h"
#include "vinfo.h"
#include "list.h"
#include "rectset.h"
#include "canvas.h"
#include "event.h"
#include "timeout.h"
#include "servant.h"


#pragma implementation "sss_ini.h"

#include "sss_ini.h"

class SSSIniElementImpl : public SSSIniElement {
	char *name_;
	mutable char *val_;
public:
	SSSIniElementImpl(char const *name, char const *val)
	{
		int lenname = strlen(name);
		int lenval = strlen(val);
		name_ = new char [lenname+1];
		val_ = new char [lenval+1];
		memcpy(name_, name, lenname+1);
		memcpy(val_, val, lenval+1);
	}
	virtual ~SSSIniElementImpl()
	{
		delete [] name_;
		delete [] val_;
	}
	virtual bool operator == (char const *name) const
	{
		return !strcasecmp(name_, name);
	}
	virtual char const *name() const { return name_; }
	virtual char const *val() const { return val_; }
	virtual void val(char const *val) const
	{
		delete [] val;
		int lenval = strlen(val);
		val_ = new char [lenval+1];
		memcpy(val_, val, lenval+1);
	}
};

class SSSIniSectionImpl : public SSSIniSection {
	char *name_;
	mutable AMList<SSSIniElement> elements_;
	mutable int numElements_;
	SSSIniElement *lookup_sub(char const *name) const
	{
		AMList<SSSIniElement>::Iterator i;
		for (i=elements_.first(); i; ++i) {
			if (*i == name) {
				return i;
			}
		}
		return NULL;
	}
public:
	SSSIniSectionImpl(char const *name) : numElements_(0)
	{
		int lenname = strlen(name);
		name_ = new char [lenname+1];
		memcpy(name_, name, lenname+1);
	}
	virtual ~SSSIniSectionImpl()
	{
		AMList<SSSIniElement>::Iterator i;

		while ((i=elements_.first())) {
			delete i.remove();
		}

		delete [] name_;
	}
	virtual bool operator == (char const *name) const
	{
		return !strcasecmp(name_, name);
	}
	virtual char const *lookup(char const *name) const
	{
		SSSIniElement *p = lookup_sub(name);
		if (p)
			return p->val();
		return NULL;
	}
	virtual void add(char const *name, char const *val) const
	{
		SSSIniElement *p = lookup_sub(name);
		if (p) {
			p->val(val);
		} else {
			p = new SSSIniElementImpl(name, val);
			elements_.insert_tail(p);
			numElements_++;
		}
	}
	virtual int num_elements() const
	{
		return numElements_;
	}
	virtual AMList<SSSIniElement> elements() const
	{
		return elements_;
	}
};

class SSSIniImpl : public SSSIni {
	mutable AMList<SSSIniSection> sections_;
	mutable int numSections_;
public:
	SSSIniImpl() : numSections_(0) { }
	virtual ~SSSIniImpl()
	{
		AMList<SSSIniSection>::Iterator i;

		while ((i=sections_.first())) {
			delete i.remove();
		}
	}
	virtual SSSIniSection const *lookup(char const *name) const
	{
		AMList<SSSIniSection>::Iterator i;
		for (i=sections_.first(); i; ++i) {
			if (*i == name) {
				return i;
			}
		}
		return NULL;
	}
	virtual SSSIniSection const *add(char const *name) const
	{
		SSSIniSection const *p = lookup(name);
		if (!p) {
			p = new SSSIniSectionImpl(name);
			sections_.insert_tail(const_cast<SSSIniSection *>(p));
			numSections_++;
		}
		return p;
	}
	virtual int num_sections() const { return numSections_; }
	virtual AMList<SSSIniSection> sections() const
	{
		return sections_;
	}
};



#define BUFSIZE 256
#define SECTION_BEGIN '['
#define SECTION_END ']'
#define SEP '='
#define COMM ';'

SSSIni *SSSIniFactory::create(char const *name)
{
	FILE *fp;
	char buf[BUFSIZE];
	SSSIni *ret = new SSSIniImpl;
	char *p, *q;
	SSSIniSection const *cursec = NULL;
	int line;

	fp = fopen(name, "r");
	if (!fp) {
		AMUSE_DPR(("cannot open %s\n", name));
		goto err;
	}

	line = 0;
	for (;;) {
		line++;
		if (!fgets(buf, sizeof(buf)-1, fp))
			break;
		buf[255] = 0;  // failsafe
		p = buf;
		while (*p && isspace(*p))
			p++;
		if ((q=strrchr(p, '\n')))
			*q = '\0';
		if ((q=strrchr(p, '\r')))
			*q = '\0';
		if ((q=strchr(p, COMM)))
			*q = '\0';

		switch (*p) {
		case '\0':
		case '\n':
			continue;
		case SECTION_BEGIN:
			if (!(q=strchr(p, SECTION_END))) {
				AMUSE_DPR((
					"error(%d): cannot find %c\n",
					line, SECTION_END));
				goto err;
			}
			p++;
			if (p==q) {
				AMUSE_DPR(("error(%d): empty section\n",
					   line));
				goto err;
			}
			*q = '\0';
			// AMUSE_DPR(("section: %s\n", p));
			cursec = ret->add(p);
			if (!cursec) {
				AMUSE_DPR((
					"error(%d): cannot create section\n",
					line));
				goto err;
			}
			break;
		default:
			if (!cursec) {
				AMUSE_DPR((
					"error(%d): not enter section\n",
					line));
				goto err;
			}
			if (!(q=strchr(p, SEP))) {
				AMUSE_DPR(("error(%d): cannot find %c\n",
					   line, SEP));
				goto err;
			}
			if (p==q) {
				AMUSE_DPR(("error(%d): empty name\n",
					   line));
				goto err;
			}
			*q = '\0'; q++;
			// AMUSE_DPR(("name: %s, val: %s\n", p, q));
			cursec->add(p, q);
		}
	}

	fclose(fp);
	return ret;
err:
	if (fp)
		fclose(fp);
	delete ret;
	return NULL;
}
