/*	$AoiMoe: x11.cc,v 1.12 2001/08/08 19:42:12 tshiozak Exp $	*/

/*
 * Copyright (c) 2001 Takuya SHIOZAKI <AoiMoe@imou.to>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */


#include <string.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include "debug.h"
#include "types.h"
#include "vinfo.h"
#include "vinfo_p.h"
#include "list.h"
#include "rectset.h"
#include "canvas.h"
#include "x11.h"

// ----------------------------------------------------------------------

class X11SimpleScreen : public X11Screen {
public:
	X11SimpleScreen(char const *dpy, int num)
	{
		display_ = XOpenDisplay(dpy);
		if (display_) {
			if (num>=0)
				number_ = num;
			else
				number_  = DefaultScreen(display_);
		}
	}
	~X11SimpleScreen()
	{
		if (*this)
			XCloseDisplay(display_);
	}
};

X11Screen *
X11ScreenFactory::create(char const *dpy, int num)
{
	X11Screen *screen;

	screen = new X11SimpleScreen(dpy, num);
	if (screen && !*screen) {
		delete screen;
		screen = NULL;
	}

	return screen;
}


// ----------------------------------------------------------------------

WH X11Drawable::get_wh() const
{
	Window dumw;
	int dumi;
	unsigned int dumu, w, h;

	XGetGeometry(get_display(), to_drawable(), &dumw, &dumi, &dumi,
		     &w, &h, &dumu, &dumu);
	return WH(w, h);
}

// ----------------------------------------------------------------------

class X11SimpleWindow : public X11Window {
	static Pos const defaultPos;
	static WH const defaultWH;
public:
	X11SimpleWindow(X11Screen const &s,
			WH const &wh,
			unsigned int borderWidth) :
		X11Window(s,
			  XCreateSimpleWindow(
				  s.to_display(),
				  RootWindow(s.to_display(), s.to_number()),
				  0, 0, wh.w(), wh.h(),
				  borderWidth,
				  s.get_black_pixel(),
				  s.get_white_pixel()))
	{
	}
	~X11SimpleWindow()
	{
		if (*this)
			XDestroyWindow(get_display(), to_window());
	}
};


X11Window *
X11WindowFactory::create_simple(X11Screen const &s,
				WH const &wh,
				unsigned int borderWidth)
{
	X11Window *win;

        win = new X11SimpleWindow(s, wh, borderWidth);
	if (win && !*win) {
		delete win;
		win = NULL;
	}

	return win;
}


// ----------------------------------------------------------------------

bool X11VisualInfo::match_depth(XPixmapFormatValues *pfv, int depth)
{
	int n, i;
	XPixmapFormatValues *p;

	p = XListPixmapFormats(get_display(), &n);
	if (p) {
		for (i=0; i<n; i++) {
			if (p[i].depth == depth) {
				memcpy(pfv, &p[i], sizeof(*pfv));
				XFree(p);
				return false;
			}
		}
		XFree(p);
	}
	return true;
}

void X11VisualInfo::set_visual_info(XVisualInfo *vi)
{
	XPixmapFormatValues pfv;

	if (match_depth(&pfv, vi->depth)) {
		AMUSE_DPR(("don't match depth\n"));
		return;
	}
	depth_ = vi->depth;
	hasAlpha_ = false;
	endian_ = ImageByteOrder(get_display())==LSBFirst?Little:Big;
	bits_per_pixel_ = pfv.bits_per_pixel;
	scanline_pad_ = pfv.scanline_pad;
	calc_shifts(&shifts_.r(), vi->red_mask);
	calc_shifts(&shifts_.g(), vi->green_mask);
	calc_shifts(&shifts_.b(), vi->blue_mask);

	AMUSE_DPR(("VisualInfo:\n"));
	AMUSE_DPR(("  depth=%d\n", depth_));
	AMUSE_DPR(("  has alpha=%s\n", hasAlpha_?"yes":"no"));
	AMUSE_DPR(("  endian=%s\n", endian_==Little?"Little":"Big"));
	AMUSE_DPR(("  bits per pixel=%d\n", bits_per_pixel_));
	AMUSE_DPR(("  scanline pad=%d\n", scanline_pad_));

	visualInfo_ = vi;
}

class X11SimpleVisualInfo : public X11VisualInfo {
public:
	X11SimpleVisualInfo(X11Screen const &s) :
		X11VisualInfo(s)
	{
		XVisualInfo tmp, *match, *optimal, *nvi;
		int i, n;

		tmp.c_class = TrueColor;
		match = XGetVisualInfo(s.to_display(),
				       VisualClassMask,
				       &tmp,
				       &n);
		if (match) {
			optimal = &match[0];
			for (i=1; i<n; i++) {
				if (match[i].depth > optimal->depth) {
					optimal = &match[i];
				}
			}
			nvi = new XVisualInfo;
			memcpy(nvi, optimal, sizeof(*visualInfo_));
			set_visual_info(nvi);
			XFree(match);
		} else {
			AMUSE_DPR(("don't match visual.\n"));
		}
	}
	~X11SimpleVisualInfo()
	{
		if (*this) {
			delete visualInfo_;
		}
		this->X11VisualInfo::~X11VisualInfo();
	}
};

X11VisualInfo *X11VisualInfoFactory::create(X11Screen const &s)
{
	X11VisualInfo *vinfo;

	vinfo = new X11SimpleVisualInfo(s);
	if (vinfo && !*vinfo) {
		delete vinfo;
		vinfo = NULL;
	}

	return vinfo;
}


// ----------------------------------------------------------------------

X11GCValues const X11GCValues::none;

class X11SimpleGC : public X11GC {
public:
	X11SimpleGC(X11Drawable const &d, X11GCValues const &v) :
		X11GC(d, XCreateGC(d.to_screen().to_display(),
				   d.to_drawable(),
				   v.to_mask(),
				   const_cast<XGCValues *>(&v.to_values())))
	{
	}
	~X11SimpleGC()
	{
		XFreeGC(get_display(), gc_);
		this->X11GC::~X11GC();
	}
};

X11GC *X11GCFactory::create(X11Drawable const &d, X11GCValues const &v)
{
	X11SimpleGC *gc;

	gc = new X11SimpleGC(d, v);
	if (gc && !*gc) {
		delete gc;
		gc = NULL;
	}

	return gc;
}



// ----------------------------------------------------------------------

class X11TrueColorCMap : public CMap {
	static bool fill_mono(uint32_t *ptr,
			      int lshift,
			      int rshift,
			      int bpp,
			      int endian)
	{
		uint32_t i;

		AMUSE_ASSERT(endian==LSBFirst || endian==MSBFirst);
		for (i=0; i<(uint32_t)elementMax; i++) {
			ptr[i] = 0;
			switch (bpp) {
			case 8:
				put8(&ptr[i], i>>rshift<<lshift);
				break;
			case 16:
				switch (endian) {
				case LSBFirst:
					put16le(&ptr[i], i>>rshift<<lshift);
					break;
				case MSBFirst:
					put16be(&ptr[i], i>>rshift<<lshift);
					break;
				}
				break;
			case 24:
				switch (endian) {
				case LSBFirst:
					put24le(&ptr[i], i>>rshift<<lshift);
					break;
				case MSBFirst:
					put24be(&ptr[i], i>>rshift<<lshift);
					break;
				}
				break;
			case 32:
				switch (endian) {
				case LSBFirst:
					put32le(&ptr[i], i>>rshift<<lshift);
					break;
				case MSBFirst:
					put32be(&ptr[i], i>>rshift<<lshift);
					break;
				}
				break;
			default:
				AMUSE_ASSERT(!"illegal bpp");
			}
		}
		return false;
	}
public:
	X11TrueColorCMap(X11VisualInfo const &vinfo) : CMap(RGBA<Cell *>())
	{
		int endian;
		if (vinfo->c_class != TrueColor) {
			AMUSE_DPR(("Visual class is not TrueColor.\n"));
			return;
		}
		bitsPerPixel_ = vinfo.bits_per_pixel();
		AMUSE_DPR(("bpp: %d\n", bitsPerPixel_));
		AMUSE_DPR(("  rmask=0x%08lx, gmask=0x%08lx, bmask=0x%08lx\n",
			   vinfo->red_mask,
			   vinfo->green_mask,
			   vinfo->blue_mask));
		if (!bitsPerPixel_)
			return;
		element_.r() = new uint32_t[elementMax*3];
		element_.g() = element_.r()+elementMax;
		element_.b() = element_.g()+elementMax;
		endian = ImageByteOrder(vinfo.to_screen().to_display());
		AMUSE_DPR(("  endian=%s\n", endian==LSBFirst?"little":"big"));
		if (fill_mono(element_.r(),
			      vinfo.shifts().r().lshift,
			      vinfo.shifts().r().rshift,
			      bitsPerPixel_,
			      endian))
			goto err;
		if (fill_mono(element_.g(),
			      vinfo.shifts().g().lshift,
			      vinfo.shifts().g().rshift,
			      bitsPerPixel_,
			      endian))
			goto err;
		if (fill_mono(element_.b(),
			      vinfo.shifts().b().lshift,
			      vinfo.shifts().b().rshift,
			      bitsPerPixel_,
			      endian))
			goto err;
			      
		return;
	err:
		delete [] element_.r();
		element_.r() = element_.g() = element_.b() = NULL;
		bitsPerPixel_ = 0;
	}
	~X11TrueColorCMap()
	{
		delete [] element_.r();
	}
	operator bool() { return !!bitsPerPixel_; }
};

CMap *CMapFactory::create(X11VisualInfo const &vinfo)
{
	X11TrueColorCMap *cmap;

	cmap = new X11TrueColorCMap(vinfo);
	if (!cmap || !*cmap) {
		delete cmap;
		cmap = NULL;
	}
	return cmap;
}


// ----------------------------------------------------------------------


class X11CanvasImplCore : public X11CanvasImpl {
	static bool match(X11Screen const &)
	{
		AMUSE_DPR(("X11Canvas: match to Core Protocol version.\n"));
		return true;
	}
	static X11CanvasImpl *factory(X11Canvas const &,
				      X11Screen const &,
				      X11Drawable const &);
protected:
	X11CanvasImplCore(X11Canvas const &c,
			  X11Screen const &s,
			  X11Drawable const &d) :
		X11CanvasImpl(c, s, d)
	{
	}
	~X11CanvasImplCore();
public:
	static MatchContext const matchContext_;
	virtual void sync(bool) const;
};

X11CanvasImpl *X11CanvasImplCore::factory(X11Canvas const &c,
					  X11Screen const &s,
					  X11Drawable const &d)
{
	X11CanvasImplCore *impl = NULL;
	X11GC *gc = NULL;
	XImage *image;

	impl = new X11CanvasImplCore(c, s, d);
	if (!impl)
		goto err0;
	gc = X11GCFactory::create(d, X11GCValues::none);
	if (!gc)
		goto err1;
	impl->gc_ = gc;

	image = XCreateImage(s.to_display(),
			     c.visual_info().to_info().visual,
			     c.visual_info().depth(),
			     ZPixmap, 0, NULL,
			     c.wh().w(), c.wh().h(),
			     c.visual_info().scanline_pad(), 0);
	if (!image)
		goto err2;
	image->data = new char[c.wh().h()*image->bytes_per_line];
	if (!image->data)
		goto err3;

	impl->image_ = image;

	AMUSE_DPR(("  canvas size=%dx%d, bytes_per_line=%u\n",
		   c.wh().w(), c.wh().h(), image->bytes_per_line));

	return impl;

err3:
	delete image;
err2:
	delete gc;
err1:
	delete impl;
err0:
	return NULL;
}

X11CanvasImplCore::~X11CanvasImplCore()
{
	delete image_->data; image_->data = NULL;
	XDestroyImage(image_);
	delete gc_;
}

void X11CanvasImplCore::sync(bool isclear) const
{
	for (CanvasRegion::Iterator i=canvas_.get_dirty(); i; ++i) {
		Rect r;
		canvas_.conv_rect(&r, i);
		XPutImage(drawable_.to_screen().to_display(),
			  drawable_.to_drawable(),
			  gc_->to_gc(),
			  image_,
			  r.x(), r.y(), r.x(), r.y(), r.w(), r.h());
		canvas_.notify_update(r);
	}
	if (isclear)
		canvas_.clear_dirty();
	canvas_.notify_sync(isclear);
}


X11CanvasImpl::MatchContext const X11CanvasImplCore::matchContext_ = {
	&X11CanvasImplCore::match, &X11CanvasImplCore::factory
};


// ----------------------------------------------------------------------

X11CanvasImpl::MatchContext const *X11CanvasImpl::matchContextList_[] = {
	&X11CanvasImplCore::matchContext_,
	NULL
};

X11CanvasImpl *X11CanvasImpl::create(X11Canvas const &cv,
				     X11Screen const &s,
				     X11Drawable const &d)
{
	MatchContext const **p;

	for (p=X11CanvasImpl::matchContextList_; *p; p++) {
		if ((*p)->match(s))
			break;
	}
	AMUSE_ASSERT(*p);
	return (*p)->factory(cv, s, d);
}


class X11SimpleCanvas : public X11Canvas {
public:
	X11SimpleCanvas(X11Drawable const &d,
			X11VisualInfo const &vinfo,
			CMap const &cmap) :
		X11Canvas(vinfo, cmap, d.get_wh())
	{
		impl_ = create(*this, vinfo.to_screen(), d);
		bytes_per_line_ = impl_->bytes_per_line();
	}
};

MDCanvas *MDCanvasFactory::create(X11Drawable const &d,
				  X11VisualInfo const &vi,
				  CMap const &cmap)
{
	X11Canvas *cv;

	cv = new X11SimpleCanvas(d, vi, cmap);

	if (cv && !*cv) {
		delete cv;
		cv = NULL;
	}

	return cv;
}
