/*	$AoiMoe: xaw.cc,v 1.12 2001/08/01 20:03:34 tshiozak Exp $	*/

/*
 * Copyright (c) 2001 Takuya SHIOZAKI <AoiMoe@imou.to>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */


#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Shell.h>
#include <X11/Xaw/Form.h>
#include <string.h>
#include <sys/time.h>
#include "types.h"
#include "debug.h"
#include "vinfo.h"
#include "list.h"
#include "rectset.h"
#include "canvas.h"
#include "event.h"
#include "timeout.h"
#include "servant.h"

#include "x11.h"
#include "xaw.h"
#include "xaw_p.h"

static XrmOptionDescRec options[] = {
	{ "-geometry", "*Shell.geometry", XrmoptionSepArg, NULL },
};

static String fallbacks[] = {
	"*mainWindow.translations: #override\\n"
	"  <Expose>: expose_proc()\\n"
	"  <BtnDown>: button_proc()\\n"
	"  <Motion>: motion_proc()\\n"
	"",
	NULL
};

static XtActionsRec actions[] = {
	{ "expose_proc", (XtActionProc)&XawServant::expose_proc },
	{ "button_proc", (XtActionProc)&XawServant::button_proc },
	{ "motion_proc", (XtActionProc)&XawServant::motion_proc },
};

static XawServant *app_servant;  // XXX

XawServant::XawServant(int *rargc, char **argv, WH const &wh) :
	opened_(false), idleNotifier_(*this)
{
	char const *prog;

	if ((prog=strrchr(argv[0], '/')))
		prog++;
	else
		prog="untitled";

	shell_ =
	    new XawWidget(
		    XtVaOpenApplication(
			    NULL, prog, options, XtNumber(options),
			    rargc, argv, fallbacks,
			    applicationShellWidgetClass,
			    //
			    NULL));
	mainWindow_ =
	    new XawWidget(
		    XtVaCreateManagedWidget(
			    "mainWindow", formWidgetClass, shell_->to_widget(),
			    //
			    XtNwidth, (Dimension)wh.w(),
			    XtNheight, (Dimension)wh.h(),
			    NULL));

	XtAppAddActions(to_appcon(), actions, XtNumber(actions));
}

int XawServant::mainloop() const
{
	XtRealizeWidget(shell_->to_widget());
	screen_ = new X11Screen(shell_->to_display(),
				XScreenNumberOfScreen(shell_->to_screen()));
	vinfo_ = X11VisualInfoFactory::create(*screen_);
	window_ = new X11Window(*screen_, mainWindow_->to_window());
	cmap_ = CMapFactory::create(*vinfo_);
	canvas_ = MDCanvasFactory::create(*window_, *vinfo_, *cmap_);
	timeout_ = new XawTimeOutNotifier(to_appcon());

	XtAppMainLoop(to_appcon());
	return 0;
}

void XawServant::expose_proc(Widget w, XEvent *event)
{
	XExposeEvent *ev = (XExposeEvent *)event;

	if (app_servant)
		app_servant->expose_cb(Rect(Pos(ev->x, ev->y),
					    WH(ev->width, ev->height)));
}

void XawServant::expose_cb(Rect const &rect)
{
	if (!opened_) {
		opened_ = true;
		startNotifier_.notify(startNotifier_);
	}
	if (canvas_) {
		canvas_->add_dirty(rect);
		canvas_->sync();
	}
}

void XawIdleNotifier::connect(ServantIdleReceiver *rv)
{
	if (isempty()) {
		XtAppAddWorkProc(
			servant_.to_appcon(), work_proc, (XtPointer)this);
	}
	connect_base(rv);
}

Boolean XawIdleNotifier::work_proc(XtPointer ptr)
{
	XawIdleNotifier *nf = (XawIdleNotifier *)ptr;

	nf->notify(*nf);

	return TRUE;
}

ServantIdleNotifier &XawServant::idle_notifier() const
{
	return idleNotifier_;
}

AMButtonNotifier &XawServant::button_notifier() const
{
	return buttonNotifier_;
}

AMMotionNotifier &XawServant::motion_notifier() const
{
	return motionNotifier_;
}

void XawServant::button_proc(Widget w, XEvent *event)
{
	XButtonEvent *ev = (XButtonEvent *)event;

	if (app_servant)
		app_servant->button_cb(Pos(ev->x, ev->y));
}

void XawServant::button_cb(Pos const &pos)
{
	buttonNotifier_.notify(buttonNotifier_, pos);
}

void XawServant::motion_proc(Widget w, XEvent *event)
{
	XMotionEvent *ev = (XMotionEvent *)event;

	if (app_servant)
		app_servant->motion_cb(Pos(ev->x, ev->y));
}

void XawServant::motion_cb(Pos const &pos)
{
	motionNotifier_.notify(motionNotifier_, pos);
}

Pos XawServant::get_cursor_pos() const
{
	int dum, x, y;
	Window root, wdum;
	Pos ret(-1, -1);

	if (XQueryPointer(screen_->to_display(), window_->to_window(),
			  &root, &wdum, &dum, &dum, &x, &y,
			  (unsigned int *)&dum)) {
		ret = Pos(x, y);
	}
	return ret;
}


// ----------------------------------------------------------------------

void XawTimeOutNotifier::timeout_callback(XtPointer p, XtIntervalId *)
{
	TimeOutReceiver *rv = (TimeOutReceiver *)p;
	ClientData *cd = (ClientData *)rv->clientDataOfTimeOutReceiver_;
	rv->callback(*cd->notifier);
	delete cd;
}

void
XawTimeOutNotifier::connect(TimeOutReceiver *rv, Tick tick)
{
	ClientData *cd = new ClientData;
	Tick cur = XawServant::xaw_current_tick(), delta;

	if (tick>cur)
		delta = tick - cur;
	else
		delta = 1;
	rv->clientDataOfTimeOutReceiver_ = (void *)cd;
	cd->notifier = this;
	cd->xtid = XtAppAddTimeOut(appcon_, delta,
				   &timeout_callback, (XtPointer)rv);
};

void XawTimeOutNotifier::disconnect(TimeOutReceiver *rv)
{
	ClientData *cd = (ClientData *)rv->clientDataOfTimeOutReceiver_;
	XtRemoveTimeOut(cd->xtid);
	delete cd;
	rv->clientDataOfTimeOutReceiver_ = NULL;
}

struct timeval XawServant::epoch;
inline void XawServant::get_epoch()
{
	if (!epoch.tv_sec) {
		gettimeofday(&epoch, NULL);
		epoch.tv_usec = 0;
	}
}

Tick XawServant::xaw_current_tick()
{
	struct timeval tv;

	get_epoch();
	gettimeofday(&tv, NULL);
	
	return (Tick)(tv.tv_sec-epoch.tv_sec)*1000 + tv.tv_usec/1000;
}

Tick XawServant::current_tick() const
{
	return XawServant::xaw_current_tick();
}

// ----------------------------------------------------------------------

Servant *ServantFactory::create(int *argc, char **argv, WH const &wh)
{
	XawServant *s;

	s = new XawServant(argc, argv, wh);
	if (s && !*s) {
		delete s;
		s = NULL;
	}
	app_servant = s; // XXX

	return s;
}
